//
// ds3502_bias.ino
//
// Bill Mah
// 
// 2023-02-22 initial release
//
//
#include <Adafruit_DS3502.h>
#include <Wire.h>
#include "DFRobot_RGBLCD.h"


Adafruit_DS3502 ds3502 = Adafruit_DS3502();
DFRobot_RGBLCD lcd(16,2);  //16 characters and 2 lines of show

/*   
    *  DS3502 address: A0 tied to Vcc = 0x28 + 1 = 0x29    
    *  
    *  Bias connections:
    *  
    * DS3502 RH to Bias supply +
    * DS3502 RL to Bias supply GND
    * 
    * DS3502 RW to gate
    * Negative Bias - connect DS3502 RH to source
    * Positive Bias - connect DS3502 RL to source
    * 
    * Digital inputs:
    *     D8: Mode (Set/Run)
    *     D7: Polarity (+/-)
    * 
    * Analog inputs:
    *     A4, A5: SDA, SCL
    *     A0: Bias low voltage
    *     A1: Bias high voltage
    *     A2: Bias voltage step
    *     A3: Step frequency
    *     A6: Measure bias supply voltage
*/


// set BIAS_SUPPLY_VOLTAGE as positive value
// cannot be greater than DS3502 limit of 15V

#define POSITIVE 1
#define NEGATIVE -1

#define RUN_MODE 1
#define SET_MODE 0

// program mode
int programMode = SET_MODE;

// Polarity input
int polarityPin = 7;
int BiasPolarity = NEGATIVE;

// Mode input
int modePin = 8;


// frequency variables
int frequencyPin = A0;
int delayValue = 0;
int lastDelayValue = 0;

// voltage to digital potentiomer
float biasSupplyVoltage = 0.0;
float voltageScaleFactor = 3.0;

// voltage step variables
float vLow = 2.0;
float vHigh = 4.0;
float vStep = 1.0;

// function declarations
int calcWiperValue(float VW, float VH);
void setParameters(void);
void displayParameters(int pmode);
float getVoltage(int pin);
float getVoltageStep(int pin);
float getInputVoltage(int pin);
void displayDelay();

// program start
void setup() {
  Serial.begin(115200);

  // initialize the LCD screen
  lcd.init();

  lcd.setCursor(0, 0); // column 0, line 0
  lcd.print("Bias Stair Wave");

  lcd.setCursor(0, 1); // column 0, line 1
  // initialize the DS3502 digital potentiomer
  if (!ds3502.begin(0x29)) {
    lcd.print("DS3502 not found");
    while (1);
  }
  
  lcd.print("Found DS3502");
  delay(3000);

  pinMode(polarityPin, INPUT);
  pinMode(modePin, INPUT);

 while (digitalRead(modePin) == LOW) {  // RUN mode
  lcd.setCursor(0, 0); 
  lcd.print("Change Mode to: "); 
  lcd.setCursor(0, 1); 
  lcd.print("    SET         "); 
  
 }
}

// program loop; runs after setup()
void loop() {
  int wiper;
  float v;
  
  // get delay for frequency adjust
  int delayValue = analogRead(A3);
  if (abs(delayValue - lastDelayValue) > 2) {
    //Serial.println(delayValue);
    lastDelayValue = delayValue;
  }   
  
  if (digitalRead(modePin) == HIGH) {  // SET mode
    if (programMode == RUN_MODE) {
      // switched from RUN_MODE to SET_MODE
      programMode = SET_MODE;
      displayParameters(programMode);

      // set bias output to maximum
      if (BiasPolarity == NEGATIVE) { 
        ds3502.setWiper(0);
       }
      else {
        ds3502.setWiper(127);
       }
    }
    
    setParameters();
  }
  else { // RUN mode
    if (programMode == SET_MODE) {
      programMode = RUN_MODE;
      displayParameters(programMode);
    }

    // generate bias staircase voltage
    // set the bias voltage limits (absolute value) and step size 
    //
    // following generates staircase voltage output
    for (v = vLow; v <= vHigh; v += vStep){ // v must be <= BIAS_SUPPLY_VOLTAGE 
      if (BiasPolarity == NEGATIVE) { 
        // RW will be relative to RH
        wiper = calcWiperValue(biasSupplyVoltage - v, biasSupplyVoltage);        
      }
      else {
        // RW will be relative to RL
        wiper = calcWiperValue(v, biasSupplyVoltage);
      }
    
      ds3502.setWiper(wiper);
      delay(lastDelayValue / 10);  // experiment with delay value for best XY plot
      //Serial.println(lastDelayValue / 10);
    }

    //displayDelay();
  } 

  
}



///
/// functions
//

void displayDelay()
{
  char string[17];
  
  sprintf(string, "delay = %d", lastDelayValue);
  lcd.clear();
  lcd.setCursor(0, 0); // column 0, line 0
  lcd.print(string);

}


void displayParameters(int pmode)
{
  /*
    display to LCD:

     programMode
     BiasPolarity
     biasSupplyVoltage;
     vLow;
     vHigh;
     vStep;
  */
  char string0[17];
  char string1[17];
  char s_vLow[8], s_vHigh[8], s_vStep[8]; 

  if (pmode == RUN_MODE) {
    // set mode string to "RUN" 
    strcpy(string0, "Run"); 
  }
  else {
    // set mode string to "SET"
    strcpy(string0,"Set");      
  }

  if (BiasPolarity == POSITIVE)
    strcat(string0, " + Bias ");
  else
    strcat(string0, " - Bias ");

  dtostrf(biasSupplyVoltage, 4, 1, string1);
  strcat(string0, string1);    
  strcat(string0, "V");
 
  
  lcd.setCursor(0, 0); // column 0, line 0
  lcd.print(string0);

  dtostrf(vLow, 4, 1, s_vLow);
  dtostrf(vHigh, 4, 1, s_vHigh);
  dtostrf(vStep, 4, 2, s_vStep);
  sprintf(string0, "L%s H%s %s", s_vLow, s_vHigh, s_vStep);
  lcd.setCursor(0, 1); // column 0, line 0
  lcd.print(string0);
  
}

// set variables based on user input
// polarityPin = 7
//     A0: Bias low voltage
//     A1: Bias high voltage
//     A2: Bias voltage step
//     A6: Measure bias supply voltage
//
float saved_BiasPolarity = 0.0;
float saved_biasSupplyVoltage = 0.0;
float saved_vLow = 0.0;
float saved_vHigh = 0.0;
float saved_vStep = 0.0;

void setParameters(void)
{
  int switchState;

  // get bias output polarity  
  if (digitalRead(polarityPin) == HIGH)
    BiasPolarity = POSITIVE;
  else
    BiasPolarity = NEGATIVE;  
  
  // get voltage to digital pot and scale to actual voltage
  biasSupplyVoltage = 3.0 * getInputVoltage(A6);
 

  // get low bias voltage (absolute value) 
  vLow = getVoltage(A0);
  if (vLow > vHigh)
    vLow = vHigh;

  // get high bias voltage (absolute value)
  vHigh = getVoltage(A1);
  if (vHigh < vLow)
    vHigh = vLow;

  // get bias voltage step
  vStep = getVoltageStep(A2);

  // update display if any parameters changedrameters();
  if (saved_BiasPolarity != BiasPolarity ||
      fabs(saved_biasSupplyVoltage - biasSupplyVoltage) > 0.1 ||
      fabs(saved_vLow - vLow) > 0.1 ||
      fabs(saved_vHigh - vHigh) > 0.1 ||
      fabs(saved_vStep - vStep) > 0.1) {
     saved_BiasPolarity = BiasPolarity;
     saved_biasSupplyVoltage = biasSupplyVoltage;
     saved_vLow = vLow;
     saved_vHigh = vHigh;
     saved_vStep = vStep;
      
     displayParameters(SET_MODE); 
  }
   
}


float getInputVoltage(int pin)
{
  return analogRead(pin) / 1023.0 * 5.0; 
}



// interpret ADC input as voltage 
// return a value between vLow and vHigh
float getVoltage(int pin)
{
  float v;
 
  v = analogRead(pin) * biasSupplyVoltage / 1023.0;

  // truncate to 0.1V
  v = (int)(v * 10.0) / 10.0;  

  return v;
}


// interpret ADC input as voltage step
// return a value between 0.1 and 2.0
// 0.1 0.25 0.5 1.0 1.25 1.5 2.0
float getVoltageStep(int pin)
{
  float vstep;
  float voltPer1024 = 2.0 / 1023.0;
 
  vstep = analogRead(pin); // value between 0 and 1023
  //Serial.println(vstep);
  
  // convert to discrete voltage step between 0.1V and 2V
  vstep *= voltPer1024;
  
  if (vstep < 0.25)
    vstep = 0.1;
  else if (vstep > 1.75)
    vstep = 2.0;
  else {  // use 0.25
    vstep /= 0.25;
    vstep = (int) vstep * 0.25;
  }

  return vstep;
}  

    




// calculate digital potentiometer wiper value
//
// VW = voltage at wiper RW (pin 7)
// VH = voltage at RHigh (pin 6)
// VL = voltage at RLow (pin 8) = 0.0
//
int calcWiperValue(float VW, float VH)
{
  // 1 wiper step = VH / 127;

  double wiperStep = VH / 127.0;
  
  return VW / wiperStep; 
  
}
